/*  Input Plugin for GameMusicGear ~
 *  Copyright (C) 2005 - 2010 gama, rockMax
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/* Default headers */

#include <pspkernel.h>
#include <pspdebug.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>

#include <ctype.h>

#include "../plugin.h"
#include "../SharedLib/libutils/utils.h"

// LINKER PATCH
extern "C" {
void *__dso_handle = NULL;
}

PSP_MODULE_INFO("IN_ADPLUG", 0, 1, 1);
PSP_NO_CREATE_MAIN_THREAD();
PSP_HEAP_SIZE_KB( 1024 * 4 );

#ifdef __cplusplus
extern "C" {
#endif
int module_start(SceSize argc, void* argp)
{
 	return 0;
}

int module_stop(int args, void *argp)
{
	return 0;
}
#ifdef __cplusplus
}
#endif

#include "adplug/adplug.h"
#include "adplug/emuopl.h"

// These are too heavy for the PSP ;(
#include "adplug/kemuopl.h"
#include "adplug/temuopl.h"
#include "adplug/realopl.h"
#include "adplug/surroundopl.h"

/* Defines */

/* End Defines */

/* Prototypes */
void init(void);
void reset(void);
int load( char *filename );
int load_rw( char *filename, char *addr, int size );
long seek( long time );
void cleanup( void );
int write_buffer( signed short *dest, int len );
void get_song_title( char *filename, char **title, int *length );
void set_tune(int tune);

void get_packet(void);
void send_packet(void);
/* End Prototypes */

/* Variables */
char db_path[256];
CAdPlugDatabase	*mydb = NULL;
Copl *opl = NULL;
CPlayer *p = NULL;

int samples_left, timer;

int dbloaded = 0;
int playing = 0;

/* Required */
int channels = 2;
int samplerate = 44100;
AFormat aformat = FMT_S16_LE;
int kbps = 128;

char title[128];
char artist[128];
char album[128];
char copyright[128];
char format[128];

long length;
int subtunes;

/* End Variables */

/** START of Plugin definitions setup */
InputPlugin itable =
{
	/* Populated by Plugin */
	"Adplug 2.2 Plugin",				/* The description of the input plugin */

	init,							/* Called when the plugin is loaded */
	reset, 						/* Called when need to update config */
	load,							/* Load the file */
	NULL,							/* Load the file from data */
	set_tune,						/* Set song subtune */
	seek,							/* Seek to the specified time */
	NULL,							/* Set the equalizer, most plugins won't be able to do this */
	cleanup,						/* Called when gmgear exit */
	write_buffer, 					/* Fill audio buffer */
	get_song_title,					/* Function to grab the title string */

	get_packet,					/* Called when need main app data packet*/
	send_packet,				/* Send data packet to main app */

	NULL,
	NULL,							/* Filled by Main App */
	NULL,							/* Filled by Main App */
	0,								/* Filled by Main App */
	LATIN,							/* Filled by Main App */
	CLIENT_PLAYER					/* Player type */
};
/** END of Plugin definitions setup */

/** Single Plugin Export */
InputPlugin *get_iplugin_info()
{
	return &itable;
}

void init(void)
{
	subtunes = 1;

	kbps = 0xFF;
	aformat = FMT_S16_LE;

	const char *token = ReadStringValue("database_path", itable.config);
	if( token ) strcpy(db_path, token);
	else strcpy(db_path, "ms0:/.gmgear/adplug.db");

	itable.in_packet->ID = NO_PACKET;
}

void reset(void)
{
	channels = ReadIntValue("channels", itable.config);
	samplerate = ReadFreqValue("samplerate", itable.config);

	strcpy(title, "");
	strcpy(album, "");
	strcpy(copyright, "");
	strcpy(artist, "");

	if( p ) {
		delete(p);
		p = NULL;
	}

	// Must reset opl with new samplerate
	if( opl ) {
		delete(opl);
		opl = NULL;
    }

	opl = new CEmuopl(samplerate, true, channels == 2);

	playing = 0;
}

int load( char *file )
{
	if( !dbloaded ) {
		mydb = new CAdPlugDatabase;
		if( !mydb ) return 0;

		mydb->load(db_path);
		CAdPlug::set_database(mydb);

		dbloaded = 1;
	}

	opl->init();
	p = CAdPlug::factory(file, opl);
	if( !p ) return 0;

	if( p->gettitle().length() ) strncpy(title, p->gettitle().c_str(), 128);
	if( p->getauthor().length() ) strncpy(artist, p->getauthor().c_str(), 128);

	length = p->songlength();
	if( length <= 0 ) length = 200000;

	subtunes  = p->getsubsongs();
	if( subtunes <= 0 ) subtunes = 1;

	p->update(); // First step
						// bits * channels * samplerate ?
    samples_left = (int)(channels * samplerate / p->getrefresh());
    timer = 0;

	playing = 1;

	strcpy(format, get_extension(file, true) );

	return 1;
}

long seek( long time )
{
	if( playing ) {
		p->seek((int)((float)timer/samplerate*1000) + time);
        timer += time*samplerate/1000;
        return time/100;
	}

    return 0;
}

void cleanup( void )
{
	if( p ) delete p;
	if( opl ) {
		delete opl;
		opl = NULL;
	}
	__psp_free_heap();
}

int write_buffer( signed short *dest, int len )
{
	if(playing)
    {
		int real_len = len>>1;
        int write_pos = 0;
        for(;;)
        {
            if (samples_left < real_len)
            {
                opl->update(&dest[write_pos],samples_left);
                write_pos += samples_left;
                real_len -= samples_left;
                if (!p->update())
                {// I probably *should* stop here? But the music often
                 // continues and they are almost all cut off to soon i think
                //    playing = 0;
                //    return write_pos;
                }
                samples_left = (int)( channels * samplerate / p->getrefresh());
            }
            else
            {
                opl->update(&dest[write_pos],real_len);
                write_pos += real_len;
                samples_left -= real_len;
                timer += real_len;
                return len;
            }
        }
    }
    else
        memset(dest, 0, len);
    return len;
}

void get_song_title( char *filename, char **title, int *length )
{
}

void set_tune(int tune)
{
	p->rewind(tune);
	if( p->gettitle().length() ) strncpy(title, p->gettitle().c_str(), 128);
	if( p->getauthor().length() ) strncpy(artist, p->getauthor().c_str(), 128);

	length = p->songlength();
	if( length <= 0 ) length = 200000;
}

void get_packet(void)
{
	switch( itable.in_packet->ID ) {
		case NO_PACKET:
		default:
			break;
	};
}

void send_packet(void)
{
	switch( itable.out_packet->ID ) {
		/* song data */
		case SONG_TITLE:
			itable.out_packet->data = title;
			break;
		case SONG_ARTIST:
			itable.out_packet->data = artist;
			break;
		case SONG_ALBUM_GAME:
			itable.out_packet->data = album;
			break;
		case SONG_COPYRIGHT:
			itable.out_packet->data = copyright;
			break;
		case SONG_FORMAT:
			itable.out_packet->data = format;
			break;
		case SONG_LENGTH:
			itable.out_packet->data = &length;
			break;
		case SONG_SUBTUNES:
			itable.out_packet->data = &subtunes;
			break;

		/* audio info */
		case AUDIO_CHANNELS:
			itable.out_packet->data = &channels;
			break;
		case AUDIO_SAMPLERATE:
			itable.out_packet->data = &samplerate;
			break;
		case AUDIO_FORMAT:
			itable.out_packet->data = &aformat;
			break;
		case AUDIO_KBPS:
			itable.out_packet->data = &kbps;
			break;

		/* extra data */
		case ATTACHED_PICTURE:
			itable.out_packet->size = 0;
			itable.out_packet->data = NULL;
			break;
		case NO_PACKET:
		default:
			itable.out_packet->size = 0;
			itable.out_packet->data = NULL;
			break;
	};
}

