/*
 *
 * mmap.c - Memory Mapper for Virtual Machine (like MSX) written by Mitsutaka Okazaki 2001.
 *
 * 2001-06-16 : version 0.00
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <assert.h>
#include <string.h>
#include "mmap.h"

#if defined(_MSC_VER)
#define INLINE __forceinline
#elif defined(__GNUC__)
#define INLINE __inline__
#else
#define INLINE static
#endif

/* for write only and read only memory */
#ifndef __SYMBIAN32__
static k_uint8 dummy_read_map[0x2000], dummy_write_map[0x2000] ;
#else
typedef struct _mmap_c_g {
  k_uint8 i_dummy_read_map[0x2000];
  k_uint8 i_dummy_write_map[0x2000];
} mmap_c_g;
extern void* _get_mmap_c(void);
# define dummy_read_map (((mmap_c_g*)_get_mmap_c())->i_dummy_read_map)
# define dummy_write_map (((mmap_c_g*)_get_mmap_c())->i_dummy_write_map)

void*
_mmap_c_initialize(void)
{
  mmap_c_g* self = malloc(sizeof(mmap_c_g));
  if (self) {
    memset(self, 0, sizeof(mmap_c_g));
  }
  return self;
}

void
_mmap_c_finalize(void* self)
{
  free(self);
}
#endif

#define FAILED(x) ((x)==NULL)

static BANK *BANK_new(k_uint32 type, k_uint8 *data)
{
  BANK *bank ;
  k_uint32 size ;

  if(type==BANK_16K) size = 0x4000 ;
  else if(type==BANK_8K) size = 0x2000 ;
  else return NULL ;
  
  if(FAILED(bank = malloc(sizeof(BANK)))) return NULL ;
  if(FAILED(bank->body = malloc(size)))
  {
    free(bank) ;
    return NULL ;
  }

  bank->type = type ;
  bank->refc = 0 ;
  bank->attr = BANK_READABLE ;
  if(data) memcpy(bank->body, data, size) ;

  return bank ;
}

static void BANK_delete(BANK *bank)
{
  if(bank)
  {
    free(bank->body) ;
    free(bank) ;
  }
}


void MMAP_set_page_attr(MMAP *mmap, k_uint32 page, k_uint32 attr)
{
  MMAP_set_bank_attr(mmap, mmap->current_slot[page], mmap->current_bank[page], attr) ;
}

void MMAP_set_bank_attr(MMAP *mmap, k_uint32 slot, k_uint32 bank, k_uint32 attr)
{
  int i ;
  assert(slot<0x10) ;
  assert(bank<0x100) ;

  if(mmap->bank[slot][bank]) mmap->bank[slot][bank]->attr = attr ;

  for(i=0;i<8;i++)
  {
    if((mmap->current_slot[i]==slot)&&(mmap->current_bank[i]==bank))
      MMAP_select_bank(mmap, i, slot, bank) ;
  }
}

void MMAP_set_bank_data(MMAP *mmap, k_uint32 slot, k_uint32 bank, k_uint32 type, k_uint8 *data)
{
  assert(slot<0x10) ;
  assert(bank<0x100) ;

  if(mmap->bank[slot][bank]) MMAP_unset_bank(mmap, slot, bank) ; 
  if(FAILED(mmap->bank[slot][bank] = BANK_new(type,data)))
  {
    assert(0) ; /* Out of Memory */
    return ; 
  }

  { 
    int i ; 
    for(i=0;i<8;i++) MMAP_select_bank(mmap, i, mmap->current_slot[i], mmap->current_bank[i]) ;
  }
}


void MMAP_unset_bank(MMAP *mmap, k_uint32 slot, k_uint32 bank)
{
  assert(slot<0x10) ;
  assert(bank<0x100) ;
  
  if(mmap->bank[slot][bank])
  {
    if(mmap->bank[slot][bank]->refc) mmap->bank[slot][bank]->refc-- ;
    else BANK_delete(mmap->bank[slot][bank]) ;
    mmap->bank[slot][bank] = NULL ;
  }

  {
    int i ;
    for(i=0;i<8;i++) MMAP_select_bank(mmap, i, mmap->current_slot[i], mmap->current_bank[i]) ;
  }
}


void MMAP_mirror_bank(MMAP *mmap, k_uint32 src_slot, k_uint32 src_bank, k_uint32 dst_slot, k_uint32 dst_bank)
{
  assert(src_slot<0x10) ;
  assert(src_bank<0x100) ;
  assert(dst_slot<0x10) ;
  assert(dst_bank<0x100) ;

  MMAP_unset_bank(mmap,dst_slot,dst_bank) ;

  mmap->bank[dst_slot][dst_bank] = mmap->bank[src_slot][dst_slot] ;
  if(mmap->bank[dst_slot][dst_bank]) mmap->bank[dst_slot][dst_bank]->refc ++ ;

  {
    int i ;
    for(i=0;i<8;i++) MMAP_select_bank(mmap, i, mmap->current_slot[i], mmap->current_bank[i]) ;
  }
}

void MMAP_select_bank(MMAP *mmap, k_uint32 page, k_uint32 slot, k_uint32 bank)
{
  assert(slot<0x10) ;
  assert(bank<0x100) ;

  /*
  if(mmap->bank[slog][bank])
  printf("BANK SELECT: PAGE%d - SLOT%d is BANK%d (%d) \n", page, slot, bank, mmap->bank[slot][bank]->attr) ;
  */
  if(!mmap->bank[slot][bank])
  {  
    mmap->readmap[page] = mmap->dummy_readmap ;
    mmap->writemap[page] = mmap->dummy_writemap ;
    mmap->current_slot[page] = slot ; 
    mmap->current_bank[page] = bank ;
  }
  else if(mmap->bank[slot][bank]->type == BANK_16K)
  {
    page&=6 ;

    if(mmap->bank[slot][bank]->attr&BANK_READABLE)
    {
      mmap->readmap[page] = mmap->bank[slot][bank]->body ;
      mmap->readmap[page+1] = mmap->bank[slot][bank]->body + 0x2000 ;
    }
    else
    {
      mmap->readmap[page] = mmap->dummy_readmap ;
      mmap->readmap[page+1] = mmap->dummy_readmap ;
    }

    if(mmap->bank[slot][bank]->attr&BANK_WRITEABLE)
    {
      mmap->writemap[page] = mmap->bank[slot][bank]->body ;
      mmap->writemap[page+1] = mmap->bank[slot][bank]->body + 0x2000 ;
    }
    else
    {
      mmap->writemap[page] = mmap->dummy_writemap ;
      mmap->writemap[page+1] = mmap->dummy_writemap ;
    }

    mmap->current_slot[page] = mmap->current_slot[page+1] = slot ; 
    mmap->current_bank[page] = mmap->current_bank[page+1] = bank ;
  }
  else if(mmap->bank[slot][bank]->type == BANK_8K)
  {
    if(mmap->bank[slot][bank]->attr&BANK_READABLE) mmap->readmap[page] = mmap->bank[slot][bank]->body ;
    else mmap->readmap[page] = mmap->dummy_readmap ;
    if(mmap->bank[slot][bank]->attr&BANK_WRITEABLE) mmap->writemap[page] = mmap->bank[slot][bank]->body ;
    else mmap->writemap[page] = mmap->dummy_writemap ;

    mmap->current_slot[page] = slot ; 
    mmap->current_bank[page] = bank ;
  }
  else
  {
    assert(0) ; 
  }
  
}

k_uint32 MMAP_read_memory(MMAP *mmap, k_uint32 adr)
{
  assert(adr<0x10000) ;
  return mmap->readmap[adr>>13][adr&0x1fff] ;
}

void MMAP_write_memory(MMAP *mmap, k_uint32 adr, k_uint32 val)
{
  assert(adr<0x10000) ;
  mmap->writemap[adr>>13][adr&0x1fff] = (k_uint8)(val&0xff) ;
}

MMAP *MMAP_new()
{
  MMAP *mmap ;
  int i ;

  if(FAILED(mmap = malloc(sizeof(MMAP)))) return NULL ;
  memset(mmap, 0, sizeof(MMAP)) ;

  for(i=0;i<8;i++) MMAP_select_bank(mmap,i,0,i) ;

  mmap->dummy_readmap = dummy_read_map;
  mmap->dummy_writemap = dummy_write_map;

  return mmap ;
}

void MMAP_delete(MMAP *mmap)
{
  int slot, bank ;
  for(slot=0; slot<0x10; slot++)
    for(bank=0; bank<0x100; bank++)
      MMAP_unset_bank(mmap,slot,bank) ;

  free(mmap) ;
}

