/*

	Audio Overload SDK

	Copyright (c) 2007-2008, R. Belmont and Richard Bannister.

	All rights reserved.

	Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

	* Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
	* Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
	* Neither the names of R. Belmont and Richard Bannister nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
	"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
	LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
	A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
	CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
	EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
	PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
	PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
	LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
	NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
	SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

// corlett.c

// Decodes file format designed by Neill Corlett (PSF, QSF, ...)

/*
 - First 3 bytes: ASCII signature: "PSF" (case sensitive)

- Next 1 byte: Version byte
  The version byte is used to determine the type of PSF file.  It does NOT
  affect the basic structure of the file in any way.

  Currently accepted version bytes are:
    0x01: Playstation (PSF1)
    0x02: Playstation 2 (PSF2)
    0x11: Saturn (SSF) [TENTATIVE]
    0x12: Dreamcast (DSF) [TENTATIVE]
    0x21: Nintendo 64 (USF) [RESERVED]
    0x41: Capcom QSound (QSF)

- Next 4 bytes: Size of reserved area (R), little-endian unsigned long

- Next 4 bytes: Compressed program length (N), little-endian unsigned long
  This is the length of the program data _after_ compression.

- Next 4 bytes: Compressed program CRC-32, little-endian unsigned long
  This is the CRC-32 of the program data _after_ compression.  Filling in
  this value is mandatory, as a PSF file may be regarded as corrupt if it
  does not match.

- Next R bytes: Reserved area.
  May be empty if R is 0 bytes.

- Next N bytes: Compressed program, in zlib compress() format.
  May be empty if N is 0 bytes.

The following data is optional and may be omitted:

- Next 5 bytes: ASCII signature: "[TAG]" (case sensitive)
  If these 5 bytes do not match, then the remainder of the file may be
  regarded as invalid and discarded.

- Remainder of file: Uncompressed ASCII tag data.
*/

#include <assert.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include "ao.h"
#include "corlett.h"

#include <zlib.h>

#define DECOMP_MAX_SIZE		((8 * 1024 * 1024) + 12)

/* :) FUNCTION TO SAVE MEMORY FROM DECOMPRESSION */

#define CHUNK 	0x8000			// 32K

void zerr(int ret)
{
    fputs("zerror: ", stderr);
    switch (ret) {
    case Z_ERRNO:
            fputs("error reading file\n", stderr);
        break;
    case Z_STREAM_ERROR:
        fputs("invalid compression level\n", stderr);
        break;
    case Z_DATA_ERROR:
        fputs("invalid or incomplete deflate data\n", stderr);
        break;
    case Z_MEM_ERROR:
        fputs("out of memory\n", stderr);
        break;
    case Z_VERSION_ERROR:
        fputs("zlib version mismatch!\n", stderr);
    }
}

SceUID get_file_from_zfile(const char *dest_name, SceUID source)
{
	SceUID dest;

	int ret;
    unsigned have;
    z_stream strm;
	unsigned char *in, *out;

	if( !(dest = sceIoOpen(dest_name, PSP_O_RDWR | PSP_O_CREAT, 0777)) ) {
		return 0;
	}

	SceOff source_pos = sceIoLseek(source, 0, PSP_SEEK_CUR);

    /* allocate inflate state */
    strm.zalloc = Z_NULL;
    strm.zfree = Z_NULL;
    strm.opaque = Z_NULL;
    strm.avail_in = 0;
    strm.next_in = Z_NULL;
    ret = inflateInit(&strm);
    if (ret != Z_OK) {
		sceIoClose(dest);
		zerr(ret);
        return 0;
	}

	// Allocate pipe bufferss
	in = (unsigned char *)malloc(CHUNK*sizeof(char));
	out = (unsigned char *)malloc(CHUNK*sizeof(char));
	if(!in || !out) {
		sceIoClose(dest);
		return 0;
	}

    /* decompress until deflate stream ends or end of file */
    do {
		if( !(strm.avail_in = sceIoRead(source,in,CHUNK)) ) break;
        /*if (ferror(source)) {
            (void)inflateEnd(&strm);
			fclose(dest);
			fseek(source, source_pos, SEEK_SET);
			zerr(Z_ERRNO);
			return NULL;
        }
        if (strm.avail_in == 0)
            break;
		*/
        strm.next_in = in;

        /* run inflate() on input until output buffer not full */
        do {
            strm.avail_out = CHUNK;
            strm.next_out = out;
            ret = inflate(&strm, Z_NO_FLUSH);
            assert(ret != Z_STREAM_ERROR);  /* state not clobbered */
            switch (ret) {
            case Z_NEED_DICT:
                ret = Z_DATA_ERROR;     /* and fall through */
            case Z_DATA_ERROR:
            case Z_MEM_ERROR:
                (void)inflateEnd(&strm);
				sceIoClose(dest);
				sceIoLseek(source, source_pos, PSP_SEEK_SET);
				free(in); free(out);
				zerr(ret);
                return 0;
            }
            have = CHUNK - strm.avail_out;
            if (sceIoWrite(dest,out,have) != have) {
                (void)inflateEnd(&strm);
				sceIoClose(dest);
				sceIoLseek(source, source_pos, PSP_SEEK_SET);
				free(in); free(out);
				zerr(Z_ERRNO);
                return 0;
            }
        } while (strm.avail_out == 0);

        /* done when inflate() says it's done */
    } while (ret != Z_STREAM_END);

	// Deallocate pipe buffers
	free(in);
	free(out);

    /* clean up and return */
    (void)inflateEnd(&strm);
	if( ret == Z_STREAM_END ) {
		/* SEEK FROM HERE? */
		sceIoLseek(source, source_pos, PSP_SEEK_SET);
		sceIoLseek(dest, 0, PSP_SEEK_SET);
		return dest;
	}

	sceIoClose(dest);
	zerr(Z_DATA_ERROR);

	return 0;
}

SceUID corlett_decode(SceUID input_fd, char *out_name, corlett_t **c)
{
	SceUID output_fd=0;
	uint32 buf[4];
	uint8  *buf8;
	uint32 res_area, comp_crc,  actual_crc;
	uint8 *decomp_dat, *tag_dec, *tag_dec_ptr;
	uLongf decomp_length, comp_length;
	uint32 input_len;

	// 32-bit data buffer
	if( !sceIoRead(input_fd, buf, 4*sizeof(uint32)) ) {
		return AO_FAIL;
	}

	buf8 = (int8 *)buf;

	// Check we have a PSF format file.
	if ((buf8[0] != 'P') || (buf8[1] != 'S') || (buf8[2] != 'F'))
	{
#ifdef DEBUG
		fprintf(stderr, "corlett_decode: Not PSF format\n");
#endif
		return AO_FAIL;
	}

	// Get our values
	res_area = LE32(buf[1]);
	comp_length = LE32(buf[2]);
	comp_crc = LE32(buf[3]);

	input_len = sceIoLseek(input_fd, 0, PSP_SEEK_END);
	sceIoLseek(input_fd, 4*sizeof(uint32), PSP_SEEK_SET);

	if (comp_length > 0)
	{
		// Check length
		if (input_len < comp_length + 16)
			return AO_FAIL;

#ifdef CRC_CHECK
		// Check CRC is correct
		// TODO - enable this ...
		actual_crc = crc32(0, (unsigned char *)&buf[4+(res_area/4)], comp_length);
		if (actual_crc != comp_crc) {
		    fprintf(stderr, "corlett_decode: CRC Error\n");
			return AO_FAIL;
		}
#endif

		if( !(output_fd = get_file_from_zfile(out_name, input_fd)) ) {
			return AO_FAIL;
		}

		// Skip res_area (?) and comp_length
		sceIoLseek(input_fd, comp_length, PSP_SEEK_CUR);
	}

	// Make structure
	*c = malloc(sizeof(corlett_t));
	memset(*c, 0, sizeof(corlett_t));

	strcpy((*c)->inf_title, "n/a");
	strcpy((*c)->inf_copy, "n/a");
	strcpy((*c)->inf_artist, "n/a");
	strcpy((*c)->inf_game, "n/a");
	strcpy((*c)->inf_year, "n/a");
	strcpy((*c)->inf_length, "n/a");
	strcpy((*c)->inf_fade, "n/a");

	// set reserved section pointer
	(*c)->res_section = &buf[4];
	(*c)->res_size = res_area;

	// Next check for tags
	input_len -= sceIoLseek(input_fd, 0, PSP_SEEK_CUR);
	if (input_len < 5)
		return output_fd;

	// printf("\n\nNew corlett: input len %d\n", input_len);

	tag_dec = (uint8 *)malloc(input_len);
	input_len = sceIoRead(input_fd, tag_dec, input_len);

	tag_dec_ptr = tag_dec;
	if ((tag_dec_ptr[0] == '[') && (tag_dec_ptr[1] == 'T') && (tag_dec_ptr[2] == 'A') && (tag_dec_ptr[3] == 'G') && (tag_dec_ptr[4] == ']'))
	{
		int tag, l, num_tags, data;

		// Tags found!
		tag_dec_ptr += 5;
		input_len -= 5;

		tag = 0;
		data = false;
		num_tags = 0;
		l = 0;
		while (input_len && (num_tags < MAX_UNKNOWN_TAGS))
		{
			if (data)
			{
				if ((*tag_dec_ptr == 0xA) || (*tag_dec_ptr == 0x00))
				{
					(*c)->tag_data[num_tags][l] = 0;
					data = false;
					num_tags++;
					l = 0;
				}
				else
				{
					(*c)->tag_data[num_tags][l++] = *tag_dec_ptr;
				}
			}
			else
			{
				if (*tag_dec_ptr == '=')
				{
					(*c)->tag_name[num_tags][l] = 0;
					l = 0;
					data = true;
				}
				else
				{
					(*c)->tag_name[num_tags][l++] = *tag_dec_ptr;
				}
			}

			tag_dec_ptr++;
			input_len--;
		}


		// Now, process that tag array into what we expect
		for (num_tags = 0; num_tags < MAX_UNKNOWN_TAGS; num_tags++)
		{
			// See if tag belongs in one of the special fields we have
			if (!strcasecmp((*c)->tag_name[num_tags], "_lib"))
			{
				strcpy((*c)->lib, (*c)->tag_data[num_tags]);
				(*c)->tag_data[num_tags][0] = 0;
				(*c)->tag_name[num_tags][0] = 0;
			}
			else if (!strncmp((*c)->tag_name[num_tags], "_lib2", 5))
			{
				strcpy((*c)->libaux[0], (*c)->tag_data[num_tags]);
				(*c)->tag_data[num_tags][0] = 0;
				(*c)->tag_name[num_tags][0] = 0;
			}
			else if (!strncmp((*c)->tag_name[num_tags], "_lib3", 5))
			{
				strcpy((*c)->libaux[1], (*c)->tag_data[num_tags]);
				(*c)->tag_data[num_tags][0] = 0;
				(*c)->tag_name[num_tags][0] = 0;
			}
			else if (!strncmp((*c)->tag_name[num_tags], "_lib4", 5))
			{
				strcpy((*c)->libaux[2], (*c)->tag_data[num_tags]);
				(*c)->tag_data[num_tags][0] = 0;
				(*c)->tag_name[num_tags][0] = 0;
			}
			else if (!strncmp((*c)->tag_name[num_tags], "_lib5", 5))
			{
				strcpy((*c)->libaux[3], (*c)->tag_data[num_tags]);
				(*c)->tag_data[num_tags][0] = 0;
				(*c)->tag_name[num_tags][0] = 0;
			}
			else if (!strncmp((*c)->tag_name[num_tags], "_lib6", 5))
			{
				strcpy((*c)->libaux[4], (*c)->tag_data[num_tags]);
				(*c)->tag_data[num_tags][0] = 0;
				(*c)->tag_name[num_tags][0] = 0;
			}
			else if (!strncmp((*c)->tag_name[num_tags], "_lib7", 5))
			{
				strcpy((*c)->libaux[5], (*c)->tag_data[num_tags]);
				(*c)->tag_data[num_tags][0] = 0;
				(*c)->tag_name[num_tags][0] = 0;
			}
			else if (!strncmp((*c)->tag_name[num_tags], "_lib8", 5))
			{
				strcpy((*c)->libaux[6], (*c)->tag_data[num_tags]);
				(*c)->tag_data[num_tags][0] = 0;
				(*c)->tag_name[num_tags][0] = 0;
			}
			else if (!strncmp((*c)->tag_name[num_tags], "_lib9", 5))
			{
				strcpy((*c)->libaux[7], (*c)->tag_data[num_tags]);
				(*c)->tag_data[num_tags][0] = 0;
				(*c)->tag_name[num_tags][0] = 0;
			}
			else if (!strncmp((*c)->tag_name[num_tags], "_refresh", 8))
			{
				strcpy((*c)->inf_refresh, (*c)->tag_data[num_tags]);
				(*c)->tag_data[num_tags][0] = 0;
				(*c)->tag_name[num_tags][0] = 0;
			}
			else if (!strncmp((*c)->tag_name[num_tags], "title", 5))
			{
				strcpy((*c)->inf_title, (*c)->tag_data[num_tags]);
				(*c)->tag_data[num_tags][0] = 0;
				(*c)->tag_name[num_tags][0] = 0;
			}
			else if (!strncmp((*c)->tag_name[num_tags], "copyright", 9))
			{
				strcpy((*c)->inf_copy, (*c)->tag_data[num_tags]);
				(*c)->tag_data[num_tags][0] = 0;
				(*c)->tag_name[num_tags][0] = 0;
			}
			else if (!strncmp((*c)->tag_name[num_tags], "artist", 6))
			{
				strcpy((*c)->inf_artist, (*c)->tag_data[num_tags]);
				(*c)->tag_data[num_tags][0] = 0;
				(*c)->tag_name[num_tags][0] = 0;
			}
			else if (!strncmp((*c)->tag_name[num_tags], "game", 4))
			{
				strcpy((*c)->inf_game, (*c)->tag_data[num_tags]);
				(*c)->tag_data[num_tags][0] = 0;
				(*c)->tag_name[num_tags][0] = 0;
			}
			else if (!strncmp((*c)->tag_name[num_tags], "year", 4))
			{
				strcpy((*c)->inf_year, (*c)->tag_data[num_tags]);
				(*c)->tag_data[num_tags][0] = 0;
				(*c)->tag_name[num_tags][0] = 0;
			}
			else if (!strncmp((*c)->tag_name[num_tags], "length", 6))
			{
				strcpy((*c)->inf_length, (*c)->tag_data[num_tags]);
				(*c)->tag_data[num_tags][0] = 0;
				(*c)->tag_name[num_tags][0] = 0;
			}
			else if (!strncmp((*c)->tag_name[num_tags], "fade", 4))
			{
				strcpy((*c)->inf_fade, (*c)->tag_data[num_tags]);
				(*c)->tag_data[num_tags][0] = 0;
				(*c)->tag_name[num_tags][0] = 0;
			}
		}
	}

	free(tag_dec);

	// Bingo
	return output_fd;
}

uint32 psfTimeToMS(char *str)
{
	int x, c=0;
	uint32 acc=0;
	char s[100];

	strncpy(s,str,100);
	s[99]=0;

	for (x=strlen(s); x>=0; x--)
	{
		if (s[x]=='.' || s[x]==',')
		{
			acc=atoi(s+x+1);
			s[x]=0;
		}
		else if (s[x]==':')
		{
			if(c==0)
			{
				acc+=atoi(s+x+1)*10;
			}
			else if(c==1)
			{
				acc+=atoi(s+x+(x?1:0))*10*60;
			}

			c++;
			s[x]=0;
		}
		else if (x==0)
		{
			if(c==0)
			{
				acc+=atoi(s+x)*10;
			}
			else if(c==1)
			{
				acc+=atoi(s+x)*10*60;
			}
			else if(c==2)
			{
				acc+=atoi(s+x)*10*60*60;
			}
		}
	}

	acc*=100;
	return(acc);
}

