/*  
   XMMS-SID - SIDPlay input plugin for X MultiMedia System (XMMS)

   Get song length from SLDB for PSID/RSID files
   
   Programmed and designed by Matti 'ccr' Hamalainen <ccr@tnsp.org> 
   (C) Copyright 1999-2007 Tecnic Software productions (TNSP)

   This program is free software; you can redistribute it and/or modify 
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*/
#include "xs_length.h"
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <assert.h>

// SUPPORT
u16 fread_be16(FILE *f)
{
    return (((u16) fgetc(f)) << 8) | ((u16) fgetc(f));
}

u32 fread_be32(FILE *f) 
{
    return (((u32) fgetc(f)) << 24) |
           (((u32) fgetc(f)) << 16) |
           (((u32) fgetc(f)) << 8) |
           ((u32)  fgetc(f));
}

u16 fread_be16_rw(char *addr)
{
	char a, b;
	memcpy(&a, addr, 1);
	memcpy(&b, addr + 1, 1);
    return (((u16) a) << 8) | ((u16) b);
}

u32 fread_be32_rw(char *addr) 
{
	char a, b, c, d;
	memcpy(&a, addr, 1);
	memcpy(&b, addr + 1, 1);
	memcpy(&c, addr + 2, 1);
	memcpy(&d, addr + 3, 1);
    return (((u32) a) << 24) |
           (((u32) b) << 16) |
           (((u32) c) << 8) |
           ((u32)  d);
}

void findnext(const char *str, size_t *pos)
{
    while (str[*pos] && isspace(str[*pos]))
        (*pos)++;
}

/* Free memory allocated for given SLDB node
 */
static void xs_sldb_node_free(sldb_node_t *node)
{
    if (node) {
        /* Nothing much to do here ... */
        free(node->lengths);
        free(node);
    }
}


/* Insert given node to db linked list
 */
static void xs_sldb_node_insert(xs_sldb_t *db, sldb_node_t *node)
{
    assert(db);

    if (db->nodes) {
        /* The first node's prev points to last node */
        LPREV = db->nodes->prev;    /* New node's prev = Previous last node */
        db->nodes->prev->next = node;    /* Previous last node's next = New node */
        db->nodes->prev = node;    /* New last node = New node */
        LNEXT = NULL;    /* But next is NULL! */
    } else {
        db->nodes = node;    /* First node ... */
        LPREV = node;    /* ... it's also last */
        LNEXT = NULL;    /* But next is NULL! */
    }
}


/* Parse a time-entry in SLDB format
 */
static int xs_sldb_gettime(char *str, size_t *pos)
{
    int result, tmp;

    /* Check if it starts with a digit */
    if (isdigit(str[*pos])) {
        /* Get minutes-field */
        result = 0;
        while (isdigit(str[*pos]))
            result = (result * 10) + (str[(*pos)++] - '0');

        result *= 60;

        /* Check the field separator char */
        if (str[*pos] == ':') {
            /* Get seconds-field */
            (*pos)++;
            tmp = 0;
            while (isdigit(str[*pos])) {
                tmp = (tmp * 10) + (str[(*pos)++] - '0');
            }

            result += tmp;
        } else
            result = -2;
    } else
        result = -1;

    /* Ignore and skip the possible attributes */
    while (str[*pos] && !isspace(str[*pos]))
        (*pos)++;

    return result;
}


/* Parse one SLDB definition line, return SLDB node
 */
sldb_node_t * xs_sldb_read_entry(char *inLine)
{
    size_t linePos;
    int i;
    u8 isOK;
    sldb_node_t *tmnode;

    /* Allocate new node */
    tmnode = (sldb_node_t *) malloc(sizeof(sldb_node_t));
    if (!tmnode) {
        fprintf(stderr, "Error allocating new node. Fatal error.\n");
        return NULL;
    }
	
	// set to zero
	memset(tmnode, 0, sizeof(sldb_node_t));

    /* Get hash value */
    linePos = 0;
    for (i = 0; i < XS_MD5HASH_LENGTH; i++, linePos += 2) {
        int tmpu;
        sscanf(&inLine[linePos], "%2x", &tmpu);
        tmnode->md5Hash[i] = tmpu;
    }
        
    /* Get playtimes */
    if (inLine[linePos] != 0) {
        if (inLine[linePos] != '=') {
            fprintf(stderr, "'=' expected on column #%d.\n", linePos);
            xs_sldb_node_free(tmnode);
            return NULL;
        } else {
            size_t tmpLen, savePos;
            
            /* First playtime is after '=' */
            savePos = ++linePos;
            tmpLen = strlen(inLine);
                        
            /* Get number of sub-tune lengths */                        
            isOK = 1;
            while ((linePos < tmpLen) && isOK) {
                findnext(inLine, &linePos);

                if (xs_sldb_gettime(inLine, &linePos) >= 0)
                    tmnode->nlengths++;
                else
                    isOK = 0;
            }
            
            /* Allocate memory for lengths */
            if (tmnode->nlengths > 0) {
                tmnode->lengths = (int *) malloc(tmnode->nlengths * sizeof(int));
                if (!tmnode->lengths) {
                    fprintf(stderr, "Could not allocate memory for node.\n");
                    xs_sldb_node_free(tmnode);
                    return NULL;
                }
				
				memset(tmnode->lengths, 0, tmnode->nlengths * sizeof(int));
            } else {
                xs_sldb_node_free(tmnode);
                return NULL;
            }
            
            /* Read lengths in */
            i = 0;
            linePos = savePos;
            isOK = 1;
            while ((linePos < tmpLen) && (i < tmnode->nlengths) && isOK) {
                int l;
                
                findnext(inLine, &linePos);

                l = xs_sldb_gettime(inLine, &linePos);
                if (l >= 0)
                    tmnode->lengths[i] = l;
                else
                    isOK = 0;

                i++;
            }

            if (!isOK) {
                xs_sldb_node_free(tmnode);
                return NULL;
            } else
                return tmnode;
        }
    }

    return NULL;
}


/* Read database to memory
 */
int xs_sldb_read(xs_sldb_t *db, const char *dbFilename)
{
    FILE *inFile;
    char inLine[XS_BUF_SIZE];
    size_t lineNum;
    sldb_node_t *tmnode;
    assert(db);

    /* Try to open the file */
    if ((inFile = fopen(dbFilename, "ra")) == NULL) {
        fprintf(stderr, "Could not open SongLengthDB '%s'\n", dbFilename);
        return -1;
    }

    /* Read and parse the data */
    lineNum = 0;

    while (fgets(inLine, XS_BUF_SIZE, inFile) != NULL) {
        size_t linePos = 0;
        lineNum++;
        
        findnext(inLine, &linePos);

        /* Check if it is datafield */
        if (isxdigit(inLine[linePos])) {
            /* Check the length of the hash */
            int hashLen;
            for (hashLen = 0; inLine[linePos] && isxdigit(inLine[linePos]); hashLen++, linePos++);

            if (hashLen != XS_MD5HASH_LENGTH_CH) {
                fprintf(stderr, "Invalid MD5-hash in SongLengthDB file '%s' line #%d!\n",
                    dbFilename, lineNum);
            } else {
                /* Parse and add node to db */
                if ((tmnode = xs_sldb_read_entry(inLine)) != NULL) {
                    xs_sldb_node_insert(db, tmnode);
                } else {
                    fprintf(stderr, "Invalid entry in SongLengthDB file '%s' line #%d!\n",
                        dbFilename, lineNum);
                }
            }
        } else if ((inLine[linePos] != ';') && (inLine[linePos] != '[') && (inLine[linePos] != 0)) {
            fprintf(stderr, "Invalid line in SongLengthDB file '%s' line #%d\n",
                dbFilename, lineNum);
        }

    }

    /* Close the file */
    fclose(inFile);

    return 0;
}

/* Compare two given MD5-hashes.
 * Return: 0 if equal
 *         negative if testHash1 < testHash2
 *         positive if testHash1 > testHash2
 */
static int xs_sldb_cmphash(xs_md5hash_t testHash1, xs_md5hash_t testHash2)
{
    int i, d;

    /* Compute difference of hashes */
    for (i = 0, d = 0; (i < XS_MD5HASH_LENGTH) && !d; i++)
        d = (testHash1[i] - testHash2[i]);

    return d;
}


/* Compare two nodes
 */
static int xs_sldb_cmp(const void *node1, const void *node2)
{
    /* We assume here that we never ever get NULL-pointers or similar */
    return xs_sldb_cmphash(
        (*(sldb_node_t **) node1)->md5Hash,
        (*(sldb_node_t **) node2)->md5Hash);
}


/* (Re)create index
 */
int xs_sldb_index(xs_sldb_t * db)
{
    sldb_node_t *pCurr;
    size_t i;
    assert(db);

    /* Free old index */
    if (db->pindex) {
        free(db->pindex);
        db->pindex = NULL;
    }

    /* Get size of db */
    pCurr = db->nodes;
    db->n = 0;
    while (pCurr) {
        db->n++;
        pCurr = pCurr->next;
    }

    /* Check number of nodes */
    if (db->n > 0) {
        /* Allocate memory for index-table */
        db->pindex = (sldb_node_t **) malloc(sizeof(sldb_node_t *) * db->n);
        if (!db->pindex)
            return -1;

        /* Get node-pointers to table */
        i = 0;
        pCurr = db->nodes;
        while (pCurr && (i < db->n)) {
            db->pindex[i++] = pCurr;
            pCurr = pCurr->next;
        }

        /* Sort the indexes */
        qsort(db->pindex, db->n, sizeof(sldb_node_t *), xs_sldb_cmp);
    }

    return 0;
}


/* Free a given song-length database
 */
void xs_sldb_free(xs_sldb_t * db)
{
    sldb_node_t *pCurr, *next;

    if (!db)
        return;

    /* Free the memory allocated for nodes */
    pCurr = db->nodes;
    while (pCurr) {
        next = pCurr->next;
        xs_sldb_node_free(pCurr);
        pCurr = next;
    }

    db->nodes = NULL;

    /* Free memory allocated for index */
    if (db->pindex) {
        free(db->pindex);
        db->pindex = NULL;
    }

    /* Free structure */
    db->n = 0;
    free(db);
}


/* Compute md5hash of given SID-file
 */
typedef struct {
    char magicID[4];    /* "PSID" / "RSID" magic identifier */
    u16 version,    /* Version number */
        dataOffset,    /* Start of actual c64 data in file */
        loadAddress,    /* Loading address */
        initAddress,    /* Initialization address */
        playAddress,    /* Play one frame */
        nSongs,        /* Number of subsongs */
        startSong;    /* Default starting song */
    u32 speed;        /* Speed */
    char sidName[32];    /* Descriptive text-fields, ASCIIZ */
    char sidAuthor[32];
    char sidCopyright[32];
} psidv1_header_t;


typedef struct {
    u16 flags;        /* Flags */
    u8 startPage, pageLength;
    u16 reserved;
} psidv2_header_t;


static int xs_get_sid_hash(const char *filename, xs_md5hash_t hash)
{
    FILE *inFile;
    xs_md5state_t inState;
    psidv1_header_t psidH;
    psidv2_header_t psidH2;
    u8 *songData;
    u8 ib8[2], i8;
    int index, result;

    /* Try to open the file */
    if ((inFile = fopen(filename, "rb")) == NULL)
        return -1;

    /* Read PSID header in */
    fread(psidH.magicID, sizeof(psidH.magicID), 1, inFile);
    if (strncmp(psidH.magicID, "PSID", 4) && strncmp(psidH.magicID, "RSID", 4)) {
        fclose(inFile);
        fprintf(stderr, "Not a PSID or RSID file '%s'\n", filename);
        return -2;
    }

    psidH.version = fread_be16(inFile);
    psidH.dataOffset = fread_be16(inFile);
    psidH.loadAddress = fread_be16(inFile);
    psidH.initAddress = fread_be16(inFile);
    psidH.playAddress = fread_be16(inFile);
    psidH.nSongs = fread_be16(inFile);
    psidH.startSong = fread_be16(inFile);
    psidH.speed = fread_be32(inFile);

    fread(psidH.sidName, sizeof(char), sizeof(psidH.sidName), inFile);
    fread(psidH.sidAuthor, sizeof(char), sizeof(psidH.sidAuthor), inFile);
    fread(psidH.sidCopyright, sizeof(char), sizeof(psidH.sidCopyright), inFile);
    
    if (feof(inFile) || ferror(inFile)) {
        fclose(inFile);
        fprintf(stderr, "Error reading SID file header from '%s'\n", filename);
        return -4;
    }
    
    /* Check if we need to load PSIDv2NG header ... */
    psidH2.flags = 0;    /* Just silence a stupid gcc warning */
    
    if (psidH.version == 2) {
        /* Yes, we need to */
        psidH2.flags = fread_be16(inFile);
        psidH2.startPage = fgetc(inFile);
        psidH2.pageLength = fgetc(inFile);
        psidH2.reserved = fread_be16(inFile);
    }

    /* Allocate buffer */
    songData = (u8 *) malloc(XS_SIDBUF_SIZE * sizeof(u8));
    if (!songData) {
        fclose(inFile);
        fprintf(stderr, "Error allocating temp data buffer for file '%s'\n", filename);
        return -3;
    }

    /* Read data to buffer */
    result = fread(songData, sizeof(u8), XS_SIDBUF_SIZE, inFile);
    fclose(inFile);

    /* Initialize and start MD5-hash calculation */
    xs_md5_init(&inState);

    if (psidH.loadAddress == 0) {
        /* Strip load address (2 first bytes) */
        xs_md5_append(&inState, &songData[2], result - 2);
    } else {
        /* Append "as is" */
        xs_md5_append(&inState, songData, result);
    }

    /* Free buffer */
    free(songData);

    /* Append header data to hash */
#define XSADDHASH(QDATAB) do {                    \
    ib8[0] = (QDATAB & 0xff);                \
    ib8[1] = (QDATAB >> 8);                    \
    xs_md5_append(&inState, (u8 *) &ib8, sizeof(ib8));    \
    } while (0)

    XSADDHASH(psidH.initAddress);
    XSADDHASH(psidH.playAddress);
    XSADDHASH(psidH.nSongs);
#undef XSADDHASH

    /* Append song speed data to hash */
    i8 = 0;
    for (index = 0; (index < psidH.nSongs) && (index < 32); index++) {
        i8 = (psidH.speed & (1 << index)) ? 60 : 0;
        xs_md5_append(&inState, &i8, sizeof(i8));
    }

    /* Rest of songs (more than 32) */
    for (index = 32; index < psidH.nSongs; index++) {
        xs_md5_append(&inState, &i8, sizeof(i8));
    }

    /* PSIDv2NG specific */
    if (psidH.version == 2) {
        /* SEE SIDPLAY HEADERS FOR INFO */
        i8 = (psidH2.flags >> 2) & 3;
        if (i8 == 2)
            xs_md5_append(&inState, &i8, sizeof(i8));
    }

    /* Calculate the hash */
    xs_md5_finish(&inState, hash);

    return 0;
}

static int xs_get_sid_hash_rw(char *addr, int size, xs_md5hash_t hash)
{
    xs_md5state_t inState;
    psidv1_header_t psidH;
    psidv2_header_t psidH2;
    u8 *songData;
    u8 ib8[2], i8;
    int index, result;
	int offset = 0;

    /* Read PSID header in */
    //fread(psidH.magicID, sizeof(psidH.magicID), 1, inFile);
	memcpy(psidH.magicID, addr, sizeof(psidH.magicID));
	offset += sizeof(psidH.magicID);
	
    if (strncmp(psidH.magicID, "PSID", 4) && strncmp(psidH.magicID, "RSID", 4)) {
        fprintf(stderr, "Not a PSID or RSID file\n");
        return -2;
    }

    psidH.version = fread_be16_rw(addr+offset);
	offset += 2;
    psidH.dataOffset = fread_be16_rw(addr+offset);
	offset += 2;
    psidH.loadAddress = fread_be16_rw(addr+offset);
	offset += 2;
    psidH.initAddress = fread_be16_rw(addr+offset);
	offset += 2;
    psidH.playAddress = fread_be16_rw(addr+offset);
	offset += 2;
    psidH.nSongs = fread_be16_rw(addr+offset);
	offset += 2;
    psidH.startSong = fread_be16_rw(addr+offset);
	offset += 2;
    psidH.speed = fread_be32_rw(addr+offset);
	offset += 4;

    //fread(psidH.sidName, sizeof(char), sizeof(psidH.sidName), inFile);
	memcpy(psidH.sidName, addr + offset, sizeof(psidH.sidName) * sizeof(char));
	offset += sizeof(psidH.sidName) * sizeof(char);
    //fread(psidH.sidAuthor, sizeof(char), sizeof(psidH.sidAuthor), inFile);
	memcpy(psidH.sidAuthor, addr + offset, sizeof(psidH.sidAuthor) * sizeof(char));
	offset += sizeof(psidH.sidAuthor) * sizeof(char);
    //fread(psidH.sidCopyright, sizeof(char), sizeof(psidH.sidCopyright), inFile);
    memcpy(psidH.sidCopyright, addr + offset, sizeof(psidH.sidCopyright) * sizeof(char));
	offset += sizeof(psidH.sidCopyright) * sizeof(char);
	
	if( offset >= size ) {
    //if (feof(inFile) || ferror(inFile)) {
        fprintf(stderr, "Error reading SID file header from address\n");
        return -4;
    }
    
    /* Check if we need to load PSIDv2NG header ... */
    psidH2.flags = 0;    /* Just silence a stupid gcc warning */
    
    if (psidH.version == 2) {
        /* Yes, we need to */
        psidH2.flags = fread_be16_rw(addr+offset);
		offset += 2;
        //psidH2.startPage = fgetc(inFile);
		memcpy(&psidH2.startPage, addr + offset, 1);
		offset ++;
        //psidH2.pageLength = fgetc(inFile);
		memcpy(&psidH2.pageLength, addr + offset, 1);
		offset ++;
        psidH2.reserved = fread_be16_rw(addr+offset);
		offset += 2;
    }

    /* Allocate buffer */
    songData = (u8 *) malloc(XS_SIDBUF_SIZE * sizeof(u8));
    if (!songData) {
        fprintf(stderr, "Error allocating temp data buffer\n");
        return -3;
    }

    /* Read data to buffer */
    //result = fread(songData, sizeof(u8), XS_SIDBUF_SIZE, inFile);
	result = sizeof(u8) * XS_SIDBUF_SIZE;
	if( (offset + result) > size ) 
		result = size - offset;
	memcpy(songData, addr + offset, result);

    /* Initialize and start MD5-hash calculation */
    xs_md5_init(&inState);

    if (psidH.loadAddress == 0) {
        /* Strip load address (2 first bytes) */
        xs_md5_append(&inState, &songData[2], result - 2);
    } else {
        /* Append "as is" */
        xs_md5_append(&inState, songData, result);
    }

    /* Free buffer */
    free(songData);

    /* Append header data to hash */
#define XSADDHASH(QDATAB) do {                    \
    ib8[0] = (QDATAB & 0xff);                \
    ib8[1] = (QDATAB >> 8);                    \
    xs_md5_append(&inState, (u8 *) &ib8, sizeof(ib8));    \
    } while (0)

    XSADDHASH(psidH.initAddress);
    XSADDHASH(psidH.playAddress);
    XSADDHASH(psidH.nSongs);
#undef XSADDHASH

    /* Append song speed data to hash */
    i8 = 0;
    for (index = 0; (index < psidH.nSongs) && (index < 32); index++) {
        i8 = (psidH.speed & (1 << index)) ? 60 : 0;
        xs_md5_append(&inState, &i8, sizeof(i8));
    }

    /* Rest of songs (more than 32) */
    for (index = 32; index < psidH.nSongs; index++) {
        xs_md5_append(&inState, &i8, sizeof(i8));
    }

    /* PSIDv2NG specific */
    if (psidH.version == 2) {
        /* SEE SIDPLAY HEADERS FOR INFO */
        i8 = (psidH2.flags >> 2) & 3;
        if (i8 == 2)
            xs_md5_append(&inState, &i8, sizeof(i8));
    }

    /* Calculate the hash */
    xs_md5_finish(&inState, hash);

    return 0;
}


/* Get node from db index via binary search
 */
sldb_node_t *xs_sldb_get(xs_sldb_t *db, const char *filename)
{
    sldb_node_t keyItem, *key, **item;

    /* Check the database pointers */
    if (!db || !db->nodes || !db->pindex)
        return NULL;

    /* Get the hash and then look up from db */
    if (xs_get_sid_hash(filename, keyItem.md5Hash) == 0) {
        key = &keyItem;
        item = bsearch(&key, db->pindex, db->n,
            sizeof(db->pindex[0]), xs_sldb_cmp);
        
        if (item)
            return *item;
        else
            return NULL;
    } else
        return NULL;
}


sldb_node_t *xs_sldb_get_rw(xs_sldb_t *db, char *addr, int size)
{
    sldb_node_t keyItem, *key, **item;

    /* Check the database pointers */
    if (!db || !db->nodes || !db->pindex)
        return NULL;

    /* Get the hash and then look up from db */
    if (xs_get_sid_hash_rw(addr, size, keyItem.md5Hash) == 0) {
        key = &keyItem;
        item = bsearch(&key, db->pindex, db->n,
            sizeof(db->pindex[0]), xs_sldb_cmp);
        
        if (item)
            return *item;
        else
            return NULL;
    } else
        return NULL;
}

