
#include <stdio.h>
#include <stdlib.h>

#include <pspkerneltypes.h>
#include <pspthreadman.h>
#include <pspkerror.h>
#include "mutex.h"

/* Create a semaphore */
semaphore *create_semaphore(unsigned int initial_value)
{
	semaphore *sem;

	sem = (semaphore *) malloc(sizeof(*sem));
	if (sem != NULL) {
		/* TODO: Figure out the limit on the maximum value. */
		sem->semid = sceKernelCreateSema("sema", 0, initial_value, 255, NULL);
		if (sem->semid < 0) {
			fprintf(stderr, "Couldn't create semaphore");
			free(sem);
			sem = NULL;
		}
	} else {
		return NULL;
	}

	return sem;
}

/* Free the semaphore */
void destroy_semaphore(semaphore *sem)
{
	if (sem != NULL) {
		if (sem->semid > 0) {
			sceKernelDeleteSema(sem->semid);
			sem->semid = 0;
		}

		free(sem);
	}
}

/* TODO: This routine is a bit overloaded.
 * If the timeout is 0 then just poll the semaphore; if it's SDL_MUTEX_MAXWAIT, pass
 * NULL to sceKernelWaitSema() so that it waits indefinitely; and if the timeout
 * is specified, convert it to microseconds. */
int sem_wait_timeout(semaphore *sem, unsigned int timeout)
{
	unsigned int *pTimeout;
    unsigned int res;

	if (sem == NULL) {
		fprintf(stderr, "Passed a NULL sem");
		return 0;
	}

	if (timeout == 0) {
		res = sceKernelPollSema(sem->semid, 1);
		if (res < 0) {
			return MUTEX_TIMEDOUT;
		}
		return 0;
	}

	if (timeout == MUTEX_MAXWAIT) {
		pTimeout = NULL;
	} else {
		timeout *= 1000;  /* Convert to microseconds. */
		pTimeout = &timeout;
	}

	res = sceKernelWaitSema(sem->semid, 1, pTimeout);
       switch (res) {
               case SCE_KERNEL_ERROR_OK:
                       return 0;
               case SCE_KERNEL_ERROR_WAIT_TIMEOUT:
                       return MUTEX_TIMEDOUT;
               default:
                       fprintf(stderr, "WaitForSingleObject() failed");
                       return -1;
	}
}

int sem_try_wait(semaphore *sem)
{
	return sem_wait_timeout(sem, 0);
}

int sem_wait(semaphore *sem)
{
	return sem_wait_timeout(sem, MUTEX_MAXWAIT);
}

/* Returns the current count of the semaphore */
unsigned int sem_value(semaphore *sem)
{
	SceKernelSemaInfo info;

	if (sem == NULL) {
		fprintf(stderr, "Passed a NULL sem");
		return 0;
	}

	if (sceKernelReferSemaStatus(sem->semid, &info) >= 0) {
		return info.currentCount;
	}

	return 0;
}

int sem_post(semaphore *sem)
{
	int res;

	if (sem == NULL) {
		fprintf(stderr, "Passed a NULL sem");
		return -1;
	}

	res = sceKernelSignalSema(sem->semid, 1);
	if (res < 0) {
		fprintf(stderr, "sceKernelSignalSema() failed");
		return -1;
	}

	return 0;
}

mutex *create_mutex(void)
{
	mutex *mut;

	/* Allocate mutex memory */
	mut = (mutex *)malloc(sizeof(*mut));
	if ( mut ) {
		/* Create the mutex semaphore, with initial value 1 */
		mut->sem = create_semaphore(1);
		mut->recursive = 0;
		mut->owner = 0;
		if ( ! mut->sem ) {
			free(mut);
			mut = NULL;
		}
	} else {
		return NULL;
	}
	return mut;
}

void destroy_mutex(mutex *mut)
{
if ( mut ) {
		if ( mut->sem ) {
			destroy_semaphore(mut->sem);
		}
		free(mut);
	}
}

int mutexp(mutex *mut)
{
	unsigned int this_thread;

	if ( mut == NULL ) {
		fprintf(stderr, "Passed a NULL mutex");
		return -1;
	}

	this_thread = (unsigned int) sceKernelGetThreadId();
	if ( mut->owner == this_thread ) {
		++mut->recursive;
	} else {
		/* The order of operations is important.
		   We set the locking thread id after we obtain the lock
		   so unlocks from other threads will fail.
		*/
		sem_wait(mut->sem);
		mut->owner = this_thread;
		mut->recursive = 0;
	}

	return 0;
}

int mutexv(mutex *mut)
{
	if ( mut == NULL ) {
		fprintf(stderr, "Passed a NULL mutex");
		return -1;
	}

	/* If we don't own the mutex, we can't unlock it */
	if ( (unsigned int) sceKernelGetThreadId() != mut->owner ) {
		fprintf(stderr, "mutex not owned by this thread");
		return -1;
	}

	if ( mut->recursive ) {
		--mut->recursive;
	} else {
		/* The order of operations is important.
		   First reset the owner so another thread doesn't lock
		   the mutex and set the ownership before we reset it,
		   then release the lock semaphore.
		 */
		mut->owner = 0;
		sem_post(mut->sem);
	}
	return 0;
}
