
/*  Input Plugin for GameMusicGear ~
 *  Copyright (C) 2005 - 2010 gama, rockMax
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/* Default headers */

#include <pspkernel.h>
#include "../plugin.h"

#include "vgmstream.h"
#include "util.h"

PSP_MODULE_INFO("IN_STREAM", 0, 1, 1);
PSP_NO_CREATE_MAIN_THREAD();
PSP_HEAP_SIZE_KB( 6 * 1024 );

#ifdef __cplusplus
extern "C" {
#endif
int module_start(SceSize argc, void* argp)
{
 	return 0;
}

int module_stop(int args, void *argp)
{
	return 0;
}
#ifdef __cplusplus
}
#endif

/* Prototypes */
void init(void);
void reset(void);
int  load( char *filename );
int  load_rw( char *filename, char *addr, int size );
long  seek( long time );
void cleanup( void );
int  write_buffer( signed short *dest, int len );
void get_song_title( char *filename, char **title, int *length );

void get_packet(void);
void send_packet(void);
/* End Prototypes */

/* Variables */
VGMSTREAM *s = NULL;
int32_t total_samples;
int32_t current_sample;

int ignore_loop;
int force_loop;
int loop_count;

int playing = 0;
int channels = 1;
int samplerate = 44100;
AFormat aformat = FMT_S16_LE;
int kbps = 128;

char title[128];
char artist[128];
char album[128];
char copyright[128];
char format[128];

long length;
int subtunes;
/* End Variables */

/** START of Plugin definitions setup */
InputPlugin itable =
{
	/* Populated by Plugin */
	"VGMStream",				/* The description of the input plugin */

	init,							/* Called when the plugin is loaded */
	reset, 							/* Called when need to update config */
	load,							/* Load the file */
	NULL,                           /* Load the file from data */
	NULL,							/* Set song subtune */
	NULL,							/* Seek to the specified time */
	NULL,							/* Set the equalizer, most plugins won't be able to do this */
	cleanup,						/* Called when gmgear exit */
	write_buffer, 					/* Fill audio buffer */
	get_song_title,					/* Function to grab the title string */

	get_packet,						/* Called when need main app data packet*/
	send_packet,					/* Send data packet to main app */

	NULL,							/* Filled by Main App */
	NULL,							/* Filled by Main App */
	NULL,							/* Filled by Main App */
	0,								/* Filled by Main App */
	LATIN,							/* Filled by Main App */
	CLIENT_PLAYER					/* Player type */
};
/** END of Plugin definitions setup */

/** Single Plugin Export */
InputPlugin *get_iplugin_info()
{
	return &itable;
}

void init(void)
{
	strcpy(format, "STREAM");
	subtunes = 1;

	kbps = 0x6C;
	aformat = FMT_S16_LE;

	itable.in_packet->ID = NO_PACKET;
}

/* CONFIG PARSER */
int read_int_value( char *setting, configType *config )
{
    unsigned int i;
    for( i = 0; i < config->n_settings; i++ )
        if( !strcmp( setting, config->setting_list[i]->name ) )
            return atoi( config->setting_list[i]->val );

    return 0;
}

int read_bool_value( char *setting, configType *config )
{
    unsigned int i;
    for( i = 0; i < config->n_settings; i++ )
		if( !strcmp( setting, config->setting_list[i]->name ) ) {
            return atoi( config->setting_list[i]->val );
		}

    return 0;
}
/* END CONFIG PARSER */

void reset(void)
{
	strcpy(title, "");
	strcpy(album, "");
	strcpy(copyright, "");
	strcpy(artist, "");

	length = 0;
	total_samples = 0;
	current_sample = 0;

	loop_count = read_int_value("loop_count", itable.config);
	force_loop = read_bool_value("force_loop", itable.config);
	ignore_loop = read_bool_value("ignore_loop", itable.config);

	// Cleanup
	if( s ) {
		close_vgmstream(s);
		s = NULL;
	}

	itable.in_packet->ID = NO_PACKET;
}

int load( char *file )
{
	s = init_vgmstream(file);
	if( !s ) return -1;

	// loop flag
	/* force only if there aren't already loop points */
    if (force_loop && !s->loop_flag) {
        /* this requires a bit more messing with the VGMSTREAM than I'm
			* comfortable with... */
        s->loop_flag = 1;
        s->loop_start_sample = 0;
        s->loop_end_sample = s->num_samples;
        s->loop_ch = calloc(s->channels,sizeof(VGMSTREAMCHANNEL));
    }

    if (ignore_loop) s->loop_flag=0;

	channels   = s->channels;
	samplerate = s->sample_rate;
	kbps = get_vgmstream_frame_size(s);
	total_samples = get_vgmstream_play_samples((double)loop_count, 0, 0, s);
	length = (total_samples*1000) / s->sample_rate;
	if( length <= 0 ) length = 200000;

	char *ext = strrchr(file, '.');
	if( ext ) {
		strcpy(format, ext+1);
	}
	else strcpy(format, "STREAM");

	playing = 1;
	return 1;
}

void cleanup( void )
{
	// Cleanup
	if( s ) close_vgmstream(s);
	__psp_free_heap();
}

int write_buffer( signed short *dest, int len )
{
	if(playing)
    {
		// Audio write func
		// have we finished decoding?
		current_sample += len / (channels << 1);
		if( current_sample >= total_samples ) {
			playing = 0;
			return 0;
		}

		render_vgmstream((sample *)dest, len / (channels << 1), s);
		return len;
    }

    return 0;
}

void get_song_title( char *filename, char **title, int *length )
{
}

void get_packet(void)
{
	switch( itable.in_packet->ID ) {
		case NO_PACKET:
		default:
			break;
	};
}

void send_packet(void)
{
	switch( itable.out_packet->ID ) {
		/* song data */
		case SONG_TITLE:
			itable.out_packet->data = title;
			break;
		case SONG_ARTIST:
			itable.out_packet->data = artist;
			break;
		case SONG_ALBUM_GAME:
			itable.out_packet->data = album;
			break;
		case SONG_COPYRIGHT:
			itable.out_packet->data = copyright;
			break;
		case SONG_FORMAT:
			itable.out_packet->data = format;
			break;
		case SONG_LENGTH:
			itable.out_packet->data = &length;
			break;
		case SONG_SUBTUNES:
			itable.out_packet->data = &subtunes;
			break;

		/* audio info */
		case AUDIO_CHANNELS:
			itable.out_packet->data = &channels;
			break;
		case AUDIO_SAMPLERATE:
			itable.out_packet->data = &samplerate;
			break;
		case AUDIO_FORMAT:
			itable.out_packet->data = &aformat;
			break;
		case AUDIO_KBPS:
			itable.out_packet->data = &kbps;
			break;

		/* extra data */
		case ATTACHED_PICTURE:
			itable.out_packet->size = 0;
			itable.out_packet->data = NULL;
			break;
		case NO_PACKET:
		default:
			itable.out_packet->size = 0;
			itable.out_packet->data = NULL;
			break;
	};
}
